Neko 1.99.7
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
opr_conv1.cu
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2026, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include <string.h>
36#include <stdlib.h>
37#include <stdio.h>
38#include "conv1_kernel.h"
39#include "elem_block_tune.h"
41#include <device/cuda/check.h>
42
43extern "C" {
44 #include <common/neko_log.h>
45}
46
47template < const int >
48int tune_conv1(void *du, void *u,
49 void *vx, void *vy, void *vz,
50 void *dx, void *dy, void *dz,
51 void *drdx, void *dsdx, void *dtdx,
52 void *drdy, void *dsdy, void *dtdy,
53 void *drdz, void *dsdz, void *dtdz,
54 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel);
55
56extern "C" {
57
61 void cuda_conv1(void *du, void *u,
62 void *vx, void *vy, void *vz,
63 void *dx, void *dy, void *dz,
64 void *drdx, void *dsdx, void *dtdx,
65 void *drdy, void *dsdy, void *dtdy,
66 void *drdz, void *dsdz, void *dtdz,
67 void *jacinv, int *nel, int *gdim, int *lx) {
68
69 static int autotune[17] = { 0 };
70 /* elements per block candidate chosen by the tuner */
71 static int autotune_eb[17] = { 0 };
72 /* chunk candidate chosen for the 1d variant */
73 static int autotune_ch[17] = { 0 };
74
75 const dim3 nthrds_1d(1024, 1, 1);
76 const dim3 nthrds_kstep((*lx), (*lx), 1);
77 const dim3 nblcks((*nel), 1, 1);
78 const cudaStream_t stream = (cudaStream_t) glb_cmd_queue;
79
80#define CASE_1D(LX, C) \
81 conv1_kernel_1d<real, LX, NEKO_CHUNKS(LX, C)> \
82 <<<nblcks, NEKO_CHUNKS_NTHRDS(LX, C), 0, stream>>> \
83 ((real *) du, (real *) u, \
84 (real *) vx, (real *) vy, (real *) vz, \
85 (real *) dx, (real *) dy, (real *) dz, \
86 (real *) drdx, (real *) dsdx, (real *) dtdx, \
87 (real *) drdy, (real *) dsdy, (real *) dtdy, \
88 (real *) drdz, (real *) dsdz, (real *) dtdz, \
89 (real *) jacinv); \
90 CUDA_CHECK(cudaGetLastError());
91
92/* Runtime dispatch onto the tuned chunk candidate */
93#define CASE_1D_SEL(LX, SEL) \
94 switch (SEL) { \
95 case 0: CASE_1D(LX, 0); break; \
96 case 1: CASE_1D(LX, 1); break; \
97 case 2: CASE_1D(LX, 2); break; \
98 default: CASE_1D(LX, 3); break; \
99 }
100
101#define CASE_KSTEP(LX, C) \
102 conv1_kernel_kstep<real, LX, NEKO_EB(LX, C)> \
103 <<<NEKO_EB_NBLCKS(*nel, LX, C), NEKO_EB_NTHRDS(LX, C), 0, stream>>> \
104 ((real *) du, (real *) u, \
105 (real *) vx, (real *) vy, (real *) vz, \
106 (real *) dx, (real *) dy, (real *) dz, \
107 (real *) drdx, (real *) dsdx, (real *) dtdx, \
108 (real *) drdy, (real *) dsdy, (real *) dtdy, \
109 (real *) drdz, (real *) dsdz, (real *) dtdz, \
110 (real *) jacinv, *nel); \
111 CUDA_CHECK(cudaGetLastError());
112
113/* Runtime dispatch onto the tuned candidate */
114#define CASE_KSTEP_SEL(LX, SEL) \
115 switch (SEL) { \
116 case 0: CASE_KSTEP(LX, 0); break; \
117 case 1: CASE_KSTEP(LX, 1); break; \
118 default: CASE_KSTEP(LX, 2); break; \
119 }
120
121#define CASE(LX) \
122 case LX: \
123 if(autotune[LX] == 0 ) { \
124 autotune[LX]=tune_conv1<LX>(du, u, \
125 vx, vy, vz, \
126 dx, dy, dz, \
127 drdx, dsdx, dtdx, \
128 drdy, dsdy, dtdy, \
129 drdz, dsdz, dtdz, \
130 jacinv, nel, gdim, lx, &autotune_eb[LX], \
131 &autotune_ch[LX]); \
132 } else if (autotune[LX] == 1 ) { \
133 CASE_1D_SEL(LX, autotune_ch[LX]); \
134 } else if (autotune[LX] == 2 ) { \
135 CASE_KSTEP_SEL(LX, autotune_eb[LX]); \
136 } \
137 break
138
139#define CASE_LARGE(LX) \
140 case LX: \
141 CASE_KSTEP(LX, 0); \
142 break
143
144
145 if ((*lx) < 11) {
146 switch(*lx) {
147 CASE(2);
148 CASE(3);
149 CASE(4);
150 CASE(5);
151 CASE(6);
152 CASE(7);
153 CASE(8);
154 CASE(9);
155 CASE(10);
156 default:
157 {
158 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
159 exit(1);
160 }
161 }
162 }
163 else {
164 switch(*lx) {
165 CASE_LARGE(11);
166 CASE_LARGE(12);
167 CASE_LARGE(13);
168 CASE_LARGE(14);
169 CASE_LARGE(15);
170 CASE_LARGE(16);
171 default:
172 {
173 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
174 exit(1);
175 }
176 }
177 }
178 }
179}
180
181template < const int LX >
182int tune_conv1(void *du, void *u,
183 void *vx, void *vy, void *vz,
184 void *dx, void *dy, void *dz,
185 void *drdx, void *dsdx, void *dtdx,
186 void *drdy, void *dsdy, void *dtdy,
187 void *drdz, void *dsdz, void *dtdz,
188 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel) {
191 int best1 = 0;
193 const int rounds = neko_tune_rounds();
194 const int iters = neko_tune_iters();
195 const int sweep = neko_eb_sweep();
196 int best = 0;
197 int retval;
198
199 for (int c = 0; c < NEKO_EB_CANDIDATES; c++) {
201 }
202 for (int c = 0; c < NEKO_CHUNKS_CANDIDATES; c++) {
204 }
205
206 const dim3 nthrds_1d(1024, 1, 1);
207 const dim3 nthrds_kstep((*lx), (*lx), 1);
208 const dim3 nblcks((*nel), 1, 1);
209 const cudaStream_t stream = (cudaStream_t) glb_cmd_queue;
210
211 char *env_value = NULL;
212 char neko_log_buf[80];
213
214 env_value=getenv("NEKO_AUTOTUNE");
215
216 sprintf(neko_log_buf, "Autotune conv1 (lx: %d)", *lx);
218
219 *eb_sel = 0;
220 *ch_sel = 0;
221
222 if(env_value) {
223 if( !strcmp(env_value,"1D") ) {
226 sprintf(neko_log_buf,"Set by env : 1 (1D, %d chunk)",
230 return 1;
231 } else if( !strcmp(env_value,"KSTEP") ) {
232 *eb_sel = neko_eb_env();
234 sprintf(neko_log_buf,"Set by env : 2 (KSTEP, %d elem/block)",
238 return 2;
239 } else {
240 sprintf(neko_log_buf, "Invalid value set for NEKO_AUTOTUNE");
242 }
243 }
244
247
248 /* Warm every variant before timing anything: each specialisation has to be
249 resident and the clocks at steady state, or whichever is timed first is
250 measured on a colder part */
251 for (int i = 0; i < NEKO_TUNE_WARMUP; i++) {
252 CASE_1D(LX, 0);
253 CASE_1D(LX, 1);
254 CASE_1D(LX, 2);
255 CASE_1D(LX, 3);
256 CASE_KSTEP(LX, 0);
257 if (sweep) {
258 CASE_KSTEP(LX, 1);
259 CASE_KSTEP(LX, 2);
260 }
261 }
262
263 /* Interleaved rounds, best time per variant */
264 for (int r = 0; r < rounds; r++) {
270 if (sweep) {
273 }
274 }
275
279 *eb_sel = best;
280 *ch_sel = best1;
281
282 if (time1[best1] < time2[best]) {
283 retval = 1;
284 } else {
285 retval = 2;
286 }
287
288 /* Leave the chosen kernel's output in place: the tuner stands in for a real
289 evaluation and the variants do not sum in the same order */
290 if (retval == 1) {
292 } else {
294 }
295
296 if (retval == 1) {
297 sprintf(neko_log_buf, "Chose : 1 (1D, %d chunk)",
299 } else {
300 sprintf(neko_log_buf, "Chose : 2 (KSTEP, %d elem/block)",
302 }
305 return retval;
306}
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdy
const int i
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ u
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ jacinv
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ vz
__global__ void const T *__restrict__ const T *__restrict__ vx
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ vy
#define NEKO_CHUNKS_CANDIDATES
Definition elem_block.h:125
#define NEKO_EB_CANDIDATES
Definition elem_block.h:63
#define NEKO_EB_SEL(LX, SEL)
Definition elem_block.h:108
#define NEKO_CHUNKS_SEL(LX, SEL)
Definition elem_block.h:147
#define NEKO_TUNE_TIME(T, LAUNCH, LX, C, ITERS)
static int neko_eb_env()
static int neko_tune_rounds()
#define NEKO_TUNE_LOG(LX, T1, T2)
#define NEKO_TUNE_INIT
#define NEKO_TUNE_BEST(T, BEST, N)
static int neko_tune_iters()
static int neko_chunks_env()
static int neko_eb_sweep()
#define NEKO_TUNE_WARMUP
void log_error(char *msg)
void log_message(char *msg)
void log_end_section()
void log_section(char *msg)
int tune_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel)
Definition opr_conv1.cu:182
#define CASE_KSTEP_SEL(LX, SEL)
#define CASE(LX)
void cuda_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx)
Definition opr_conv1.cu:61
#define CASE_1D_SEL(LX, SEL)
#define CASE_KSTEP(LX, C)
#define CASE_LARGE(LX)
#define CASE_1D(LX, C)